////////////////////////////////////////////////////////////////////////////////
//
// Fangorn Wizards Lab Extension Library v1.35
//  Fangorn Wizards Lab 1998 - 2003
// This package is FREEWARE :)
//

//  :
// 20  2003 -   ShlOBJ,
// -    
//   Delphi

//  :
// 

unit FWOpenDirectory;

interface

uses
  Windows, Classes, Controls, FWUtils, {$IFDEF VER130} DsgnIntf
  {$ELSE} DesignIntf {$ENDIF};

type
  PSHItemID = ^TSHItemID;
  _SHITEMID = record
    cb: Word;
    abID: array[0..0] of Byte;
  end;
  TSHItemID = _SHITEMID;

  PItemIDList = ^TItemIDList;
  _ITEMIDLIST = record
     mkid: TSHItemID;
   end;
  TItemIDList = _ITEMIDLIST;

  BFFCALLBACK = function(Wnd: HWND; uMsg: UINT; lParam, lpData: LPARAM): Integer stdcall;
  TFNBFFCallBack = type BFFCALLBACK;

  _browseinfoA = record
    hwndOwner: HWND;
    pidlRoot: PItemIDList;
    pszDisplayName: PAnsiChar;
    lpszTitle: PAnsiChar;
    ulFlags: UINT;
    lpfn: TFNBFFCallBack;
    lParam: LPARAM;          
    iImage: Integer;
  end;
  TBrowseInfoA = _browseinfoA;
  TBrowseInfo = TBrowseInfoA;

  TFWOpenDirStyle = (odsReturnOnlyFSDirs, odsDontGoBelowDomain, odsStatusText,
    odsReturnFSAncestors, odsEditBox, odsValidate, odsNewDialogStyle,
    odsBrowseIncludeUrls, odsBrowseForComputer, odsBrowseForPrinter,
    odsBrowseIncludeFiles, odsShareable, odsUseNewUI);
  TFWOpenDirStyles = set of TFWOpenDirStyle;

  TFWOpenDirectory = class(TComponent)
  private
    FAbout: String;
    FTitle: String;
    FDirName: String;
    FStyle: TFWOpenDirStyles;
    FStyleArray: array [0..12] of Integer;
    FHandle: THandle;
  protected
    procedure Init; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    function Execute: Boolean;
    property DirName: String read FDirName;
  published
    property About: String read FAbout write FAbout;
    property Title: String read FTitle write FTitle;
    property Style: TFWOpenDirStyles read FStyle write FStyle default [odsReturnOnlyFSDirs];
  end;

const
  BIF_RETURNONLYFSDIRS   = $0001;  
  BIF_DONTGOBELOWDOMAIN  = $0002;  
  BIF_STATUSTEXT         = $0004;
  BIF_RETURNFSANCESTORS  = $0008;
  BIF_EDITBOX            = $0010;
  BIF_VALIDATE           = $0020;  
  BIF_NEWDIALOGSTYLE     = $0040;
  BIF_BROWSEINCLUDEURLS  = $0080;
  BIF_BROWSEFORCOMPUTER  = $1000;  
  BIF_BROWSEFORPRINTER   = $2000;  
  BIF_BROWSEINCLUDEFILES = $4000;
  BIF_SHAREABLE          = $8000;
  BIF_USENEWUI = BIF_NEWDIALOGSTYLE or BIF_EDITBOX;

  function SHBrowseForFolder(var lpbi: TBrowseInfo): PItemIDList; stdcall;
  function SHGetPathFromIDList(pidl: PItemIDList; pszPath: PChar): BOOL; stdcall;
  function SHBrowseForFolder; external 'shell32.dll' name 'SHBrowseForFolderA';
  function SHGetPathFromIDList; external 'shell32.dll' name 'SHGetPathFromIDListA';
  procedure Register;
  
implementation

procedure Register;
begin
  RegisterComponents('Fangorn Wizards Lab', [TFWOpenDirectory]);
  RegisterPropertyEditor(TypeInfo(String), TFWOpenDirectory, 'About', TAboutPropertyEditor);
end;

{ TFWOpenDirectory }

constructor TFWOpenDirectory.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FHandle := TWinControl(AOwner).Handle;
  FStyle := [odsReturnOnlyFSDirs];
  Init;
end;

function TFWOpenDirectory.Execute: Boolean;
var
  lpItemID: PItemIDList;
  BrowseInfo: TBrowseInfo;
  DisplayName: array[0..MAX_PATH] of Char;
  TempPath: array[0..MAX_PATH] of Char;
  I, Flags: Integer;
begin
  TempPath[0] := #0;
  FillChar(BrowseInfo, sizeof(TBrowseInfo), #0);
  BrowseInfo.hwndOwner := FHandle;
  BrowseInfo.pszDisplayName := @DisplayName;
  BrowseInfo.lpszTitle := PAnsiChar(FTitle);

  Flags := 0;
  for I:= 0 to 12 do
    if (TFWOpenDirStyle(I) in FStyle) then 
    	Flags:= Flags or FStyleArray[I];
  BrowseInfo.ulFlags := Flags;
  lpItemID := SHBrowseForFolder(BrowseInfo);
  if Assigned(lpItemId) then 
  begin
    Result := True;
    SHGetPathFromIDList(lpItemID, TempPath);
    GlobalFreePtr(lpItemID);
  end 
  else 
    Result := False;
  FDirName := String(TempPath);
end;

procedure TFWOpenDirectory.Init;
begin
  FStyleArray[0] := BIF_RETURNONLYFSDIRS;
  FStyleArray[1] := BIF_DONTGOBELOWDOMAIN;
  FStyleArray[2] := BIF_STATUSTEXT;
  FStyleArray[3] := BIF_RETURNFSANCESTORS;
  FStyleArray[4] := BIF_EDITBOX;
  FStyleArray[5] := BIF_VALIDATE;
  FStyleArray[6] := BIF_NEWDIALOGSTYLE;
  FStyleArray[7] := BIF_BROWSEINCLUDEURLS;
  FStyleArray[8] := BIF_BROWSEFORCOMPUTER;
  FStyleArray[9] := BIF_BROWSEFORPRINTER;
  FStyleArray[10] := BIF_BROWSEINCLUDEFILES ;
  FStyleArray[11] := BIF_SHAREABLE;
  FStyleArray[12] := BIF_USENEWUI;
end;

end.
 