////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Unit Name : GSStorage
//  * Purpose   :      (Com Storage)
//  * Author    :  
//  * Copyright :   2001 - 2004 .
//  * Version   : 1.07
//  ****************************************************************************
//

unit GSStorage;

interface

uses
  Windows,
  Classes,
  ActiveX,
  AxCtrls,
  SysUtils;

const
  CLRF = #13#10;

type
  TStatStgArray = array of TStatStg;
  TGSStorageEnum = packed record
    Count: Cardinal;
    Enum: TStatStgArray;
  end;

  TGSStorageCursors = class;
  TGSCustomStorage = class;
  TGSStorage = class;

  TGSStorageCursor = class(TCollectionItem)
  private
    FLockCount: Integer;
    FID: ShortString;
    FStorageInterface: IStorage;
    FSubStorage: TGSStorageCursors;
    FMode: Longint;
    FOwner: TGSCustomStorage;
    FParent: TGSStorageCursor;
    procedure SetParam(const AOwner: TGSCustomStorage;
      AParent: TGSStorageCursor);
    function GetPath: String;
    function GetLock: Boolean;
    function GetLockCount: Integer;
    procedure Lock;
    procedure UnLock;
    property Locked: Boolean read GetLock;
  protected
    procedure Garbage;
    function ExistsObject(const AName: String; const AType: Longint;
      out Exists: Boolean): Boolean;
    function GetName: String;
  public
    function Backward(var AStorage: TGSStorageCursor): Boolean;
    constructor Create(Collection: TCollection); override;
    function CreateStream(const AName: String): HRESULT;
    function CreateStorage(const AName: String;
      var AStorage: TGSStorageCursor): HRESULT;
    function Copy(const AName: String;
      const AStorage: TGSStorageCursor): HRESULT;
    function DeleteStream(const AName: String): HRESULT;
    function DeleteStorage(const AName: String): HRESULT;
    destructor Destroy; override;
    function Enumerate(out AData: TGSStorageEnum): HRESULT;
    function MoveTo(const AName: String;
      const AStorage: TGSStorageCursor): HRESULT;
    function OpenStorage(const AName: String;
      var AStorage: TGSStorageCursor): HRESULT;
    function StreamExists(const AName: String): Boolean;
    function StorageExists(const AName: String): Boolean;
    function ReadStream(const AName: String; out AStream: TMemoryStream): HRESULT;
    function Rename(const AName, ANewName: String): HRESULT;
    function WriteStream(const AName: String; const AStream: TMemoryStream): HRESULT;
    procedure Release;
    property Path: String read GetPath;
  end;

  TGSStorageCursors = class(TCollection)
  private
    function GetItem(Index: Integer): TGSStorageCursor;
    procedure SetItem(Index: Integer; const Value: TGSStorageCursor);
    function Add: TGSStorageCursor;
  protected
    function FindStorageByName(
      const AName: String; out Index: Cardinal): Boolean;
  public
    constructor Create;
    property Items[Index: Integer]: TGSStorageCursor read GetItem write SetItem;
  end;

  TGSCustomStorage = class(TPersistent)
  private
    FRootStorage: TGSStorageCursor;
    FMode: Longint;
    FFileName: String;
    FCanCreate: Boolean;
  protected
    function OpenFileEx(const AFileName: String;
      const CanCreate: Boolean; const AMode: LongInt;
      out AStorage: TGSStorageCursor): HRESULT;
    procedure Garbage;
  public
    function CreateCursor: TGSStorageCursor;
    destructor Destroy; override;
    function ForceStorage(APath: String; var AStorage: TGSStorageCursor): HRESULT;
    function OpenFileReadOnly(const AFileName: String;
      out AStorage: TGSStorageCursor): HRESULT;
    function OpenFile(const AFileName: String;
      const CanCreate: Boolean; out AStorage: TGSStorageCursor): HRESULT;
    function OpenStorage(APath: String; var AStorage: TGSStorageCursor): HRESULT;
    procedure CloseFile;
    function IsBussy: Boolean;
    function ReConnect(var AStorage: TGSStorage;
      var ACursor: TGSStorageCursor): HRESULT;
  end;

  TGSStorage = class(TGSCustomStorage);

 { TGInfoStorage = class(TGSCustomStorage)
  public
    function WriteInteger
  end; }

implementation

{ TGSStorage }

//     Storage
// =============================================================================
procedure TGSCustomStorage.CloseFile;
begin
  if Assigned(FRootStorage) then
    FreeAndNil(FRootStorage);
end;

function TGSCustomStorage.CreateCursor: TGSStorageCursor;
begin
  if FRootStorage = nil then
  begin
    Result := nil;
    Exit;
  end;
  Result := FRootStorage;
end;

destructor TGSCustomStorage.Destroy;
begin
  if Assigned(FRootStorage) then
    FRootStorage.Free;
  inherited;
end;


function TGSCustomStorage.IsBussy: Boolean;
var
  LocCount: Integer;
begin
  LocCount := 0;
  Inc(LocCount, FRootStorage.GetLockCount);
  Result := LocCount > 1;
end;


function TGSCustomStorage.ReConnect(var AStorage: TGSStorage;
  var ACursor: TGSStorageCursor): HRESULT;
begin
  AStorage.CloseFile;
  Result := AStorage.OpenFileEx(FFileName, FCanCreate, FMode, ACursor);
end;


//   ForceDirectory
// =============================================================================
function TGSCustomStorage.ForceStorage(APath: String;
  var AStorage: TGSStorageCursor): HRESULT;
var
  S: TStringList;
  I: Integer;
begin
  Result := S_FALSE;
  if APath = '' then
    raise Exception.Create(SysErrorMessage(ERROR_INVALID_PARAMETER));
  S := TStringList.Create;
  try
    AStorage := FRootStorage;
    if AStorage.GetName + '\' <> Copy(APath, 1, Length(AStorage.GetName) + 1) then
    begin
      Result := ERROR_PATH_NOT_FOUND;
      Exit;
    end;
    Delete(APath, 1, Length(AStorage.GetName) + 1);
    S.Text := StringReplace(APath, '\', CLRF, [rfReplaceAll]);
    for I := 0 to S.Count - 1 do
    begin
      if AStorage.StorageExists(S.Strings[I]) then
        Result := AStorage.OpenStorage(S.Strings[I], AStorage)
      else
        Result := AStorage.CreateStorage(S.Strings[I], AStorage);
      if Result <> S_OK then
      begin
        FRootStorage.Garbage;
        Exit;
      end;
    end;
  finally
    S.Free;
  end;
end;      

procedure TGSCustomStorage.Garbage;
begin
  if FRootStorage <> nil then
    FRootStorage.Garbage;
end;


//    (   )
// =============================================================================
function TGSCustomStorage.OpenFile(const AFileName: String;
  const CanCreate: Boolean; out AStorage: TGSStorageCursor): HRESULT;
begin
  Result := OpenFileEx(AFileName, CanCreate,
    STGM_READWRITE or STGM_SHARE_EXCLUSIVE, AStorage);
end;


//    
// =============================================================================
function TGSCustomStorage.OpenFileEx(const AFileName: String;
  const CanCreate: Boolean; const AMode: LongInt;
  out AStorage: TGSStorageCursor): HRESULT;
begin
  FMode := AMode;
  FCanCreate := CanCreate;
  FFileName := AFileName;
  CloseFile;
  FRootStorage := TGSStorageCursor.Create(nil);
  if CanCreate then
  begin
    AStorage := FRootStorage;
    AStorage.SetParam(Self, nil);
    Result := StgCreateDocfile(StringToOleStr(AFileName), FMode or STGM_CREATE, 0,
      AStorage.FStorageInterface);
    AStorage.Lock;
  end
  else
  begin
    if not FileExists(AFileName) then
    begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;
    Result := StgIsStorageFile(StringToOleStr(AFileName));
    if Result <> S_OK then Exit;
    AStorage := FRootStorage;
    AStorage.SetParam(Self, nil);
    Result := StgOpenStorage(StringToOleStr(AFileName),
      nil, FMode, nil, 0, AStorage.FStorageInterface);
    AStorage.Lock;
  end;
end;


function TGSCustomStorage.OpenStorage(APath: String;
  var AStorage: TGSStorageCursor): HRESULT;
var
  S: TStringList;
  I: Integer;
begin
  Result := S_FALSE;
  if APath = '' then
    raise Exception.Create(SysErrorMessage(ERROR_INVALID_PARAMETER));
  S := TStringList.Create;
  try
    AStorage := FRootStorage;
    if AStorage.GetName + '\' <> Copy(APath, 1, Length(AStorage.GetName) + 1) then
    begin
      Result := ERROR_PATH_NOT_FOUND;
      Exit;
    end;
    Delete(APath, 1, Length(AStorage.GetName) + 1);
    S.Text := StringReplace(APath, '\', CLRF, [rfReplaceAll]);
    for I := 0 to S.Count - 1 do
      if AStorage.StorageExists(S.Strings[I]) then
      begin
        if AStorage.OpenStorage(S.Strings[I], AStorage) <> S_OK then Exit
      end
      else
        raise Exception.Create(SysErrorMessage(ERROR_PATH_NOT_FOUND));
    Result := S_OK;
  finally
    S.Free;
  end;
end;


//      
// =============================================================================
function TGSCustomStorage.OpenFileReadOnly(
  const AFileName: String; out AStorage: TGSStorageCursor): HRESULT;
begin
  Result := OpenFileEx(AFileName, False, STGM_READWRITE or STGM_SHARE_EXCLUSIVE,
    AStorage);
end;

{ TGSStorageCursors }

function TGSStorageCursors.Add: TGSStorageCursor;
begin
  Result := TGSStorageCursor(inherited Add);
end;

constructor TGSStorageCursors.Create;
begin
  inherited Create(TGSStorageCursor);
end;


//    Storage  
// =============================================================================
function TGSStorageCursors.FindStorageByName(
  const AName: String; out Index: Cardinal): Boolean;
var
  I: Cardinal;
begin
  Result := False;
  Index := 0;
  if Count = 0 then Exit;
  for I := 0 to Count - 1 do
    if Items[I].GetName = AName then
    begin
      Result := True;
      Index := I;
      Break;
    end;
end;

function TGSStorageCursors.GetItem(Index: Integer): TGSStorageCursor;
begin
  Result := TGSStorageCursor(inherited GetItem(Index));
end;

procedure TGSStorageCursors.SetItem(Index: Integer;
  const Value: TGSStorageCursor);
begin
  inherited SetItem(Index, Value);
end;

{ TGSStorageCursor }


//
// =============================================================================
function TGSStorageCursor.Backward(var AStorage: TGSStorageCursor): Boolean;
begin
  Result := FParent <> nil;
  if Result then
  begin
    UnLock;
    AStorage := FParent;
    AStorage.Lock;
    if not Locked then
      Free;
  end;
end;

function TGSStorageCursor.Copy(const AName: String;
  const AStorage: TGSStorageCursor): HRESULT;
begin
  if AStorage = nil then
    raise EComponentError.Create(SysErrorMessage(E_INVALIDARG));
  if FStorageInterface = nil then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  //Result := FStorageInterface.CopyTo(0, nil, nil, AStorage.FStorageInterface);
  Result := FStorageInterface.MoveElementTo(StringToOleStr(AName),
    AStorage.FStorageInterface, StringToOleStr(AName), STGMOVE_COPY);
end;

constructor TGSStorageCursor.Create(Collection: TCollection);
var
  ID: TGUID;
begin
  inherited;
  FSubStorage := TGSStorageCursors.Create;
  FLockCount := 0;
  if CreateGUID(ID) <> S_OK then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  FID := GUIDToString(ID);
end;

function TGSStorageCursor.CreateStorage(const AName: String;
  var AStorage: TGSStorageCursor): HRESULT;
var
  Index: Cardinal;
begin
  if AStorage =  nil then
    raise EComponentError.Create(SysErrorMessage(E_INVALIDARG));
  if FSubStorage.FindStorageByName(AName, Index) then
  begin
    Result := ERROR_ALREADY_EXISTS;
    Exit;
  end;
  AStorage := FSubStorage.Add;
  AStorage.SetParam(FOwner, Self);
  Result := FStorageInterface.CreateStorage(StringToOleStr(AName),
    FMode, 0, 0, AStorage.FStorageInterface);
  AStorage.Lock;
end;

function TGSStorageCursor.CreateStream(const AName: String): HRESULT;
var
  TmpStream:IStream;
begin
  if FStorageInterface = nil then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  Result := FStorageInterface.CreateStream(StringToOleStr(AName),
    FMode, 0, 0, TmpStream);
end;

function TGSStorageCursor.DeleteStorage(const AName: String): HRESULT;
var
  Index: Cardinal;
begin
  if FStorageInterface = nil then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  if FSubStorage.FindStorageByName(AName, Index) then
  begin
    if FSubStorage.Items[Index].Locked then
     raise EComponentError.Create(SysErrorMessage(ERROR_ACCESS_DENIED));
    FSubStorage.Delete(Index);
  end;
  Result := FStorageInterface.DestroyElement(StringToOleStr(AName));
end;

function TGSStorageCursor.DeleteStream(const AName: String): HRESULT;
begin
  if FStorageInterface = nil then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  Result := FStorageInterface.DestroyElement(StringToOleStr(AName));
end;

destructor TGSStorageCursor.Destroy;
begin
  FStorageInterface := nil;
  FreeAndNil(FSubStorage);
  inherited;
end;

function TGSStorageCursor.Enumerate(out AData: TGSStorageEnum): HRESULT;
var
  Enum: IEnumStatStg;
  TmpElement: TStatStg;
  I: Cardinal;
begin
  if FStorageInterface = nil then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  Result := FStorageInterface.EnumElements(0, nil, 0, Enum);
  if Result = S_OK then
  begin
    AData.Count := 0;
    while Enum.Next(1, TmpElement, nil) = S_OK do Inc(AData.Count);
    SetLength(AData.Enum, AData.Count);
    Enum.Reset;
    I := 0;
    while I < AData.Count do
    begin
      Enum.Next(1, AData.Enum[I], nil);
      //if AData.Enum[I].pwcsName = '' then Dec(I);
      Inc(I);
    end;
  end;
end;

function TGSStorageCursor.ExistsObject(const AName: String; const AType: Longint;
  out Exists: Boolean): Boolean;
var
  I: Integer;
  Data: TGSStorageEnum;
begin
  Result := Enumerate(Data) = S_OK;
  Exists := False;
  if Result then
    for I := 0 to Data.Count - 1 do
      if Data.Enum[I].pwcsName = AName then
      begin
        Exists := Data.Enum[I].dwType = AType;
        Exit;
      end;
end;

procedure TGSStorageCursor.Garbage;
var
  I: Integer;
begin
  if FSubStorage = nil then Exit;
  if FSubStorage.Count > 0 then
    for I := 0 to FSubStorage.Count - 1 do
    begin
      FSubStorage.Items[I].Garbage;
      if not FSubStorage.Items[I].Locked then
        FSubStorage.Items[I].Free;
    end;
end;

function TGSStorageCursor.GetLock: Boolean;
var
  I: Integer;
begin
  Result := FLockCount > 0;
  if not Result then
    if FSubStorage.Count > 0 then
      for I := 0 to FSubStorage.Count - 1 do
        if FSubStorage.Items[I].Locked then
        begin
          Result := True;
          Break;
        end;
end;

function TGSStorageCursor.GetLockCount: Integer;
var
  I: Integer;
begin
  Result := FLockCount;
  if FSubStorage.Count > 0 then
    for I := 0 to FSubStorage.Count - 1 do
      Inc(Result, FSubStorage.Items[I].GetLockCount);
end;

function TGSStorageCursor.GetName: String;
var
  Error: HRESULT;
  StatStg: TStatStg;
begin
  Result := '';
  if FStorageInterface = nil then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  Error := FStorageInterface.Stat(StatStg, STATFLAG_DEFAULT);
  if Error <> S_OK then
    raise Exception.Create(SysErrorMessage(Error));
  Result := StatStg.pwcsName;
end;

function TGSStorageCursor.GetPath: String;
begin
  if FParent = nil then
    Result := GetName
  else
    Result := FParent.GetPath + '\' + GetName;
end;

procedure TGSStorageCursor.Lock;
begin
  Inc(FLockCount);
end;

function TGSStorageCursor.MoveTo(const AName: String;
  const AStorage: TGSStorageCursor): HRESULT;
var
  Index: Cardinal;
begin
  if AStorage = nil then
    raise EComponentError.Create(SysErrorMessage(E_INVALIDARG));
  if FStorageInterface = nil then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  if FSubStorage.FindStorageByName(AName, Index) then
    Result := E_ACCESSDENIED
  else
    Result := FStorageInterface.MoveElementTo(StringToOleStr(AName),
        AStorage.FStorageInterface, StringToOleStr(AName), STGMOVE_MOVE);
end;

function TGSStorageCursor.OpenStorage(const AName: String;
  var AStorage: TGSStorageCursor): HRESULT;
var
  Index: Cardinal;
begin
  if AStorage = nil then
    raise EComponentError.Create(SysErrorMessage(E_INVALIDARG));
  if fSubStorage.FindStorageByName(AName, Index) then
  begin
    Result := S_OK;
    AStorage := FSubStorage.Items[Index];
    AStorage.Lock;
    Exit;
  end;
  if AStorage <> nil then
    if AStorage.FID = FID then UnLock;
  AStorage := FSubStorage.Add;
  AStorage.SetParam(FOwner, Self);
  Result := FStorageInterface.OpenStorage(StringToOleStr(AName),
    nil, FMode, nil, 0, AStorage.FStorageInterface);
  AStorage.Lock;
end;

function TGSStorageCursor.ReadStream(const AName: String;
  out AStream: TMemoryStream): HRESULT;
var
  TmpStream:IStream;
  OS:TOleStream;
begin
  if FStorageInterface = nil then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  if StreamExists(AName) then
    Result := FStorageInterface.OpenStream(StringToOleStr(AName),
      nil, FMode, 0, TmpStream)
  else
    raise EComponentError.Create(SysErrorMessage(ERROR_PATH_NOT_FOUND));
  if Result = S_OK then
  begin
    AStream := TMemoryStream.Create;
    OS := TOleStream.Create(TmpStream);
    try
      OS.Position := 0;
      AStream.CopyFrom(OS, OS.Size);
      AStream.Position := 0;
    finally
      OS.Free;
      TmpStream := nil;
    end;
  end;
end;

procedure TGSStorageCursor.Release;
begin
  if FID = FOwner.FRootStorage.FID then Exit;
  Free;
  FOwner.Garbage;
end;

function TGSStorageCursor.Rename(const AName, ANewName: String): HRESULT;
begin
  if FStorageInterface = nil then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  Garbage;
  Result := FStorageInterface.RenameElement(StringToOleStr(AName),
    StringToOleStr(ANewName));
end;

procedure TGSStorageCursor.SetParam(const AOwner: TGSCustomStorage;
  AParent: TGSStorageCursor);
begin
  FMode := AOwner.FMode;
  FOwner := AOwner;
  FParent := AParent;
end;

function TGSStorageCursor.StorageExists(const AName: String): Boolean;
var
  Exists: Boolean;
begin
  Result := False;
  if ExistsObject(AName, STGTY_STORAGE, Exists) then
    Result := Exists;
end;

function TGSStorageCursor.StreamExists(const AName: String): Boolean;
var
  Exists: Boolean;
begin
  Result := False;
  if ExistsObject(AName, STGTY_STREAM, Exists) then
    Result := Exists;
end;

procedure TGSStorageCursor.UnLock;
begin
  Dec(FLockCount);
end;

function TGSStorageCursor.WriteStream(const AName: String;
  const AStream: TMemoryStream): HRESULT;
var
  TmpStream:IStream;
  OS:TOleStream;
begin
  if FStorageInterface = nil then
    raise EComponentError.Create(SysErrorMessage(E_UNEXPECTED));
  if StreamExists(AName) then
    Result := FStorageInterface.OpenStream(StringToOleStr(AName),
      nil, FMode, 0, TmpStream)
  else
    Result := FStorageInterface.CreateStream(StringToOleStr(AName),
      FMode, 0, 0, TmpStream);
  if Result = S_OK then
  begin
    OS := TOleStream.Create(TmpStream);
    try
      AStream.Position := 0;
      OS.CopyFrom(AStream, AStream.Size);
    finally
      OS.Free;
    end;
  end;
end;


end.

